'''

arXivSpider 0.4

Created on 20.01.2010

@author: christian.staudt @ ira.uka.de

Copyright 2010 ITI Wagner, Karlsruhe Institute of Technology

This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.


'''

import urllib
from xml.dom import minidom
import datetime
import os.path
import sys


class ArxivEntry(object):
	
	pass

class IdMap(object):
	
	def __init__(self):
		self.dictionary = dict()
		self.id = 0
		self.hits = 0
	
	def get(self, name):
		if name in self.dictionary:
			self.hits += 1
			return self.dictionary[name]
		else:
			self.id += 1
			self.dictionary[name] = self.id
			return self.id
		
class LineBuffer(object):
	
	"""Buffers a number of lines before writing them to a file"""
	
	def __init__(self, filename, capacity=10):
		self.capacity = capacity
		self.counter = 0
		self.lines = []
		self.file = open(filename, 'a')
		
	def flush(self):
		for line in self.lines:
			try:
				self.file.write(line)
			except UnicodeEncodeError:
				# TODO: fix unicode problem
				print "unicode problem"
		self.lines = []
	
	def write(self, line):
		self.lines.append(line)
		self.counter += 1
		if self.counter == self.capacity:
			self.flush()
			self.counter = 0
	
	def close(self):
		self.flush()
		self.file.close()
			
	
class NotInArxivException(Exception):
	"""Raised if there is no paper with the given arxiv Identifier"""
	def __init__(self, value):
		self.parameter = value
	def __str__(self):
		return repr(self.parameter)
	
class EmptyFeedException(Exception):
	"""Raised if there is no entry in the result feed"""
	def __init__(self, value):
		self.parameter = value
	def __str__(self):
		return repr(self.parameter)
	
class ArxivParserException(Exception):
	"""Raised if something goes wrong that prevents the parser from retrieving the data"""
	def __init__(self, xmlNode):
		self.parameter = xmlNode
	def __str__(self):
		return repr(self.parameter)

		

class ArxivParser(object):
	
	def __init__(self):
		self.entryNode = None # the xml node containting the current entry from the query result
		
	def reset(self):
		self.entryNode = None
		
	
	def parse2(self, feedData):
		hitCounter = 0
		
		xmlDoc = minidom.parseString(feedData)
		feedNode = xmlDoc.firstChild # 'feed'
		entryNodes = feedNode.getElementsByTagName('entry')
		#DEBUG
		print "%d entries in feed" % len(entryNodes)
		#DEBUG
		if len(entryNodes) > 0:
			for entryNode in entryNodes:
				self.entryNode = entryNode # ?
				if self.isEmpty():
					print "empty"
					pass
				else:
					hitCounter +=1
					# create entry object
					arxivEntry = ArxivEntry()
					try:
						arxivEntry.id = self.getPaperId()
					except:
						arxivEntry.id = None
					try:
						arxivEntry.date = self.getDate()
					except:
						arxivEntry.date = None
					try:
						arxivEntry.authors = self.getAuthorNames()
					except:
						arxivEntry.authors = None
					try:
						arxivEntry.categories = self.getCategories()
					except:
						arxivEntry.categories = None
					yield arxivEntry
			print "%s entries parsed" % hitCounter
		else:
			raise EmptyFeedException("%d entries in feed" % len(entryNodes))
	
	def getDate(self):
		"""return a string with the day of publication"""
		# get publication date
		try:
			publishedNode = self.entryNode.getElementsByTagName('published')[0] # there should be only 1
			dateString =  publishedNode.firstChild.nodeValue
			return dateString
		except IndexError:
			raise ArxivParserException("no field 'published' found")
			
	
	def getPaperId(self):
		"""return a string with the id of the publication"""
		# get publication date
		idNode = self.entryNode.getElementsByTagName('id')[0] # there should be only 1
		idString =  idNode.firstChild.nodeValue
		return idString
	
	def getTitle(self):
		"""return a string with the title of the publication"""
		# get publication title
		titleNode = self.entryNode.getElementsByTagName('title')[0] # there should be only 1
		titleString =  titleNode.firstChild.nodeValue
		return titleString
	
	def getSingleField(self, name):
		try:
			node = self.entryNode.getElementsByTagName(name)[0] # there should be only 1
			content =  node.firstChild.nodeValue
			return content
		except:
			raise ArxivParserException(self.entryNode)
		
	
	def getAuthorNames(self):
		"""return a list of author names"""
		authorNodes = self.entryNode.getElementsByTagName('author')
		for authorNode in authorNodes:	
			nameNode = authorNode.childNodes[1]
			authorName = nameNode.firstChild.nodeValue
			yield authorName
			
	def getCategories(self):
		"""Return a list of categories for the entry"""
		categoryNodes = self.entryNode.getElementsByTagName("category")
		for categoryNode in categoryNodes:
			category = categoryNode.attributes["term"].value
			yield category
			
	def isEmpty(self):
		try:
			self.getTitle()
			return False
		except Exception:
			# every nonempty entry has at least a title
			return True
	
	def isError(self, entryNode):
		error = False
		if self.getTitle(entryNode) == "Error":
			error = True
		return error
	
	

	

class ArxivSpider(object):
	
	# constants
	MAXID = 9999
	MINYEAR = 7
	MAXYEAR = 10
	
	QUERYURL = 'http://export.arxiv.org/api/query'

	FIELD_SEPARATOR = ":: "
	LIST_SEPARATOR = ", "
	
	def __init__(self, writer, filters=[]):
		self.author2id = IdMap()
		self.parser = ArxivParser()
		self.filters = filters
		
		self.writer = writer
		self.entryCounter = 0
		
		
			
	def generateIdPrefixes(self, minYear=7, minMonth=4, maxYear=10, maxMonth=1):
		done = False
		
		for y in range(minYear, maxYear+1):
			for m in range(1,13):
				if (y == minYear and m < minMonth):
					pass
				elif (y == maxYear and m > maxMonth):
					done = True
					break
				else:
					yield str(y).rjust(2,'0') + str(m).rjust(2,'0')
			if done:
				break

			
	def generateIdSuffixes(self, batchSize=100, padding=4):
		idRange = range(1, self.MAXID+1)
		while idRange: 
			ids, idRange = idRange[:batchSize], idRange[batchSize:]
			idStrings = [str(id).rjust(padding,'0') for id in ids]
			yield idStrings
					
	def generateIds(self):
		"""
		"""
		for idP in self.generateIdPrefixes():
			for idS in self.generateIdSuffixes():
				arxivId = idP + "." + idS
				try:
					self.queryId(arxivId)
				except:
					suffixFail = True
					

		
	def queryId(self,id):
		url = self.QUERYURL + '?id_list=' + id
		feedData = urllib.urlopen(url).read()
		return feedData

	def queryIds(self,ids):
		"""query a list of ids"""
		idList = ",".join(ids)
		# DEBUG
		print "querying: " + ids[0] + " to " + ids[-1]
		#DEBUG
		url = self.QUERYURL + '?id_list=' + idList
		feedData = urllib.urlopen(url).read()
		return feedData
	
	
	def writeEntry(self, entry):
		authorsString = self.LIST_SEPARATOR.join(map(str, entry.authors))
		categoriesString = self.LIST_SEPARATOR.join(map(str, entry.categories))
		# there might be rouge semicola in the categories
#		if self.FIELD_SEPARATOR in categoriesString:
#			#DEBUG
#			print categoriesString
#			#DEBUG
#			categoriesString.replace(self.FIELD_SEPARATOR, self.LIST_SEPARATOR)
			
		lineData = [entry.id, entry.date, categoriesString, authorsString]
		line = self.FIELD_SEPARATOR.join(map(str, lineData)) 
		line = line.replace("\n","")
		line = line + "\n"
		self.writer.write(line)
		self.entryCounter += 1
		
	def isCollaboration(self, idList):
		return len(idList) > 1
	
	def cleanup(self):
		self.entryCounter = 0
		self.writeId2Author()
		self.writer.close()
		
	def exit(self):
		for filter in self.filters:
			print filter.getStatistics()
		print "%d entries parsed" % self.entryCounter
		self.cleanup()
		sys.exit()
		
	def prepareEntry(self, arxivEntry):	
		"""Make adjustments to the data read from the feed"""	
		#remove url and version number from id
		arxivEntry.id = arxivEntry.id.split("abs/")[1].split("v")[0]
		
		# remove hours, minutes,... from date
#		arxivEntry.date = arxivEntry.date.split("T")[0]
		# map the author names to ids
		arxivEntry.authors = [self.author2id.get(name) for name in arxivEntry.authors]
		
		return arxivEntry
	
	def passedFilters(self, arxivEntry):
		"""entry has passed the given filters"""
		passed = True
		for filter in self.filters:
			passed = passed and filter(arxivEntry)
			if not passed:
				return False
		return True

		
#	def arxivIdCrawl(self, batchSize=100):
#		# TODO: incomplete
#		for idPrefix in self.generateIdPrefixes():
#			for idSuffixBatch in spider.generateIdSuffixes(batchSize):
#				idBatch = [idPrefix + "." + str(idSuffix) for idSuffix in idSuffixBatch]
#				try:
#					feedData = self.queryIds(idBatch)
#					for arxivEntry in self.parser.parse2(feedData):
#						arxivEntry = self.prepareEntry(arxivEntry)
#						if self.passFilter(arxivEntry):
#							self.writeEntry(arxivEntry)
#				except NotInArxivException:
#					break

			
								
	def searchQuery(self, key, value, start, maxResults):
		"""Interface for the search_query method with paging"""
		url = self.QUERYURL + "?search_query=%s:%s&start=%d&max_results=%d" % (key, value, start, maxResults)
		print "querying: " + url
		feedData = urllib.urlopen(url).read()
		return feedData


	def searchCrawl(self, key, values, batchSize, maxEntries=10e9):
		localCounter = 0 # 
		for value in values:
			i = 0
			while True:
				try:
					#DEBUG
					print "fetching results %d to %d" % (i * batchSize+1, (i+1) * batchSize)
					#DEBUG
					feedData = self.searchQuery(key, value, start=i*batchSize, maxResults=batchSize)
					i += 1
					for arxivEntry in self.parser.parse2(feedData):
							arxivEntry = self.prepareEntry(arxivEntry)
							if self.passedFilters(arxivEntry):
								self.writeEntry(arxivEntry)
								localCounter += 1
					if self.entryCounter >= maxEntries:
						print "reached maximum number of parsed entries: exiting"
						self.exit()
					print "%d entries stored" % localCounter
					localCounter = 0
				except EmptyFeedException:
					print "Query returned empty feed for %s:%s. trying next value" % (key,value)
					break
		print "done with the list of values, exiting"
		self.exit()
	
	def writeId2Author(self):
		self.writer.write(">>>>>Id2Author\n")
		id2Author = dict([[v,k] for k,v in self.author2id.dictionary.items()])
		for id, author in id2Author.items():
			self.writer.write(self.FIELD_SEPARATOR.join((str(id), author)) + "\n")

## filters
class ArxivFilter(object):
	
	def __init__(self):
		self.negativeCounter = 0
		
	def keep(self):
		return True
	
	def discard(self):
		self.negativeCounter += 1
		return False
	
	def getStatistics(self):
		return "%s filtered out %d negatives" % (self.__class__.__name__, self.negativeCounter)


class CollaborationFilter(ArxivFilter):
	
	"""Leaves only entries with more than one author"""
	
	def __init__(self):
		ArxivFilter.__init__(self)

	
	def __call__(self, arxivEntry):
		if len(arxivEntry.authors) > 1:
			return self.keep()
		else:
			return self.discard()


class CategoryFilter(ArxivFilter):
	
	def __init__(self, *categories):
		ArxivFilter.__init__(self)
		self.categories = categories
	
	def __call__(self, arxivEntry):
		for category in self.categories:
			if category not in arxivEntry.categories:
				return False
		return True
		

class DuplicateFilter(ArxivFilter):
	
	"""Filters out duplicate entries using the arxiv identifier"""
	
	def __init__(self):
		ArxivFilter.__init__(self)
		self.idMap = dict()
	
	def __call__(self, arxivEntry):
		if arxivEntry.id in self.idMap:
			return self.discard()
		else:
			self.idMap[arxivEntry.id] = True
			return self.keep()




if __name__ == '__main__':
	# make filename with timestamp
	currentTime = datetime.datetime.now()
	fileName = "arxiv_" + str(currentTime).split(".")[0].replace(":","-").replace(" ","_") + ".txt"
	print fileName
	outDir = "out" # directory for output files
	filePath = os.path.join(outDir, fileName)
	#create spider
	spider = ArxivSpider(writer=LineBuffer(filePath), filters=[CollaborationFilter(), DuplicateFilter()])
	#spider.arxivIdCrawl(batchSize=500)
	
	#cond mat all
	#spider.searchCrawl(key="cat", values=["cond-mat","cond-mat.other","cond-mat.dis-nn", "cond-mat.mtrl-sci", "cond-mat.mes-hall", "cond-mat.quant-gas", "cond-mat.soft", "cond-mat.stat-mech", "cond-mat.str-el", "cond-mat.supr-con"], batchSize=1000, maxEntries=10e6)
	
	#cond mat soft
	#spider.searchCrawl(key="cat", values=["cond-mat.soft"], batchSize=1000, maxEntries=10e6)
	
	#q-bio all
	#spider.searchCrawl(key="cat", values=["q-bio","q-bio.BM","q-bio.CB","q-bio.GN","q-bio.MN","q-bio.NC","q-bio.OT","q-bio.PE","q-bio.QM","q-bio.SC","q-bio.TO"], batchSize=1000, maxEntries=10e6)
	
	#quantumt physics
	#spider.searchCrawl(key="cat", values=["quant-ph"], batchSize=1000, maxEntries=10e6)
	
	#astrophysics all
	# spider.searchCrawl(key="cat", values=["astro-ph", "astro-ph.CO", "astro-ph.EP", "astro-ph.GA", "astro-ph.HE", "astro-ph.IM", "astro-ph.SR"], batchSize=1000, maxEntries=10e6)
	
	#nonlinear systems all
	#spider.searchCrawl(key="cat", values=["nlin", "nlin.AO", "nlin.CG", "nlin.CD", "nlin.SI", "nlin.PS"], batchSize=1000, maxEntries=10e6)
	
	#hep all	
	#spider.searchCrawl(key="cat", values=["hep-ex", "hep-lat", "hep-ph", "hep-th"], batchSize=1000, maxEntries=10e6)
	
	#statistics all
	#spider.searchCrawl(key="cat", values=["stat", "stat.AP", "stat.CO", "stat.ML", "stat.ME", "stat.TH", "math.ST"], batchSize=1000, maxEntries=10e6)
	
	#mathematical physics
	#spider.searchCrawl(key="cat", values=["math-ph"], batchSize=1000, maxEntries=10e6)
	
	#quantitative finance
	#spider.searchCrawl(key="cat", values=["q-fin", "q-fin.CP", "q-fin.GN", "q-fin.PM", "q-fin.PR", "q-fin.RM", "q-fin.ST", "q-finTR", ], batchSize=1000, maxEntries=10e6)
	
	csSubCategories = ["AI", "AR", "CL", "CC", "CE", "CG", "GT", "CV", "CY", "CR",
					 "DB", "DL", "DM", "DC", "FL", "GL", "GR", "HC", "IR", "IT", "LG",
					  "LO", "MS", "MA", "MM", "NI", "NE", "NA", "OS", "OH", "PF", "PL",
					   "RO", "SE", "SD", "SC"]
	csCategories = ["cs." + sub for sub in csSubCategories] # computer science
	
	#cs all
	#spider.searchCrawl(key="cat", values=csCategories, batchSize=1000, maxEntries=10e6)
	
	#mathematics all	
	spider.searchCrawl(key="cat", values=["math", "math.AG", "math.AT", "math.AP", "math.CT", "math.CA", "math.CO", "math.AC", "math.CV", "math.DG", "math.DS", "math.FA", "math.GM", "math.GN", "math.GT", "math.GR", "math.HO", "math.IT", "math.KT", "math.LO", "math.MP", "math.MG", "math.NT", "math.NA", "math.OA", "math.OC", "math.PR", "math.QA", "math.RT", "math.RA", "math.SP", "math.ST", "math.SG", "math.",], batchSize=1000, maxEntries=10e6)
	

